/*
 * Please cite:
 * Zen, G., Ricci, E.
 * Earth Mover’s Prototypes: a Convex Learning Approach for Discovering Activity Patterns in Dynamic Scenes.
 * IEEE Conference on Computer Vision and Pattern Recognition (CVPR), 2011.
 * http://disi.unitn.it/~zen/emd.html
 
 This file is part of EMP (Earth Mover's Prototypes).

 EMP is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 EMP is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with EMP.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
//
#include <glpk.h>

#include "emp.h"
//----------------------------------------------------------------------------------------------------------------------

EMP::EMP() {

    N = 0;
    D = 0;
    Np = 0;
    gamma = 0;
    //---
    data_matrix = NULL;
    idx_cluster_sort = NULL;
    adj_matrix = NULL;
    clustered_data_matrix = NULL;
    prototype_idx = NULL;
}
//----------------------------------------------------------------------------------------------------------------------

EMP::EMP(parameters& param) {

    N = param.N;
    D = param.D;
    Np = param.Np;
    gamma = param.gamma;
    //
    init();
}
//----------------------------------------------------------------------------------------------------------------------

EMP::~EMP() {

    delete_data();
}
//----------------------------------------------------------------------------------------------------------------------

void EMP::init() {

    if (N <= 0 || D <= 0) {
        fprintf(stderr, "\nError in EMP::init(): N=%d D=%d not valid.\n", N, D);
        exit(1);
    }
    //---
    data_matrix = new double*[N];
    for (int i = 0; i < N; i++) {
        data_matrix[i] = new double[D];
        memset(data_matrix[i], 0, D * sizeof (double));
    }
    //---
    idx_cluster_sort = new int[D];
    for (int i = 0; i < D; ++i) {
        idx_cluster_sort[i] = i;
    }
    //---
    adj_matrix = new int*[N];
    for (int i = 0; i < N; i++) {
        adj_matrix[i] = new int[N];
        for (int j = 0; j < N; j++) {
            adj_matrix[i][j] = 0;
        }
    }
    //-------------------------------

    // res data
    n_prot = 0;
    //---
    clustered_data_matrix = new double*[N];
    for (int i = 0; i < N; i++) {
        clustered_data_matrix[i] = new double[D];
        memset(clustered_data_matrix[i], 0, D * sizeof (double));
    }
    //---
    prototype_idx = new int[N];
    for (int i = 0; i < N; ++i)
        prototype_idx[i] = -1;
    //--
}
//----------------------------------------------------------------------------------------------------------------------

void EMP::reset_out_data() {

    n_prot = 0;
    //---
    for (int i = 0; i < N; ++i)
        prototype_idx[i] = -1;
    //---
    for (int i = 0; i < N; i++) {
        memset(clustered_data_matrix[i], 0, D * sizeof (double));
    }
    //---
}
//----------------------------------------------------------------------------------------------------------------------

int EMP::load_data_matrix_from_file(parameters& param) {

    double *sum_data_matrix = new double[N];

    char filename[300];
    sprintf(filename, "%s/%s", param.output_dir, param.file_data);
    FILE *RT = fopen(filename, "r");
    if (RT == NULL) {
        fprintf(stderr, "ERROR: Impossible to open file \'%s\' for reading\n", filename);
        exit(1);
    }
    fprintf(stderr, "Loading data matrix from file: \'%s\'... \n", filename);

    for (int ii = 0; ii < N; ii++) {
        sum_data_matrix[ii] = 0.0;
        fscanf(RT, "%*d ");
        //fprintf(stderr, "%d) ", ii);

        for (int jj = 0; jj < D; jj++) {
            double val;
            fscanf(RT, "%lf ", &val);
            data_matrix[ii][jj] = val;
            sum_data_matrix[ii] += data_matrix[ii][jj];
            //fprintf(stderr, "%.2f ", data_matrix[ii][jj]);
        }
        //fprintf(stderr, "\n");
        fscanf(RT, "\n");
    }
    fclose(RT);

    // normalize histograms to sum 1
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < D; jj++) {
            if (sum_data_matrix[ii] > 0) {
                data_matrix[ii][jj] /= sum_data_matrix[ii];
            } else {
                data_matrix[ii][jj] = 1.0 / (double) D;
            }
            // fprintf(stderr, "%.2f ", data_matrix[ii][jj]);
        }
        //fprintf(stderr, "\n");
    }
    delete[] sum_data_matrix;
    fprintf(stderr, "data matrix rows are normalized to sum 1. ");
    return 0;
}
//----------------------------------------------------------------------------------------------------------------------

int EMP::sort_data_matrix(parameters& param) {

    if (strcmp(param.clustering_type, "L1") == 0) {
        fprintf(stderr, "\nNo sorting for case L1. ");
        return 0;
    } else if (strcmp(param.clustering_type, "L1") != 0 && !param.loadfilesort) {
        fprintf(stderr, "\n###Warning: using clustering %s with no file for sorting specified.###\n", param.clustering_type);
        return 0;
    }

    char filename_sorting[200];
    sprintf(filename_sorting, "%s/%s", param.output_dir, param.file_sort);

    load_sorting_from_file(filename_sorting);

    double** data_matrix_cp = new double*[N];
    for (int i = 0; i < N; i++) {
        data_matrix_cp[i] = new double[D];
        memcpy(data_matrix_cp[i], data_matrix[i], D * sizeof (double));
    }

    // sort
    for (int i = 0; i < D; i++) {
        int k = idx_cluster_sort[i];
        // copy data column
        for (int j = 0; j < N; ++j) {
            data_matrix[j][i] = data_matrix_cp[j][k];
        }
    }
    //---
    for (int i = 0; i < N; i++)
        delete[] data_matrix_cp[i];
    delete[] data_matrix_cp;

    return 0;
}

/* *************************************************************************************************** */

int EMP::load_sorting_from_file(char* filename) {

    FILE* pf = fopen(filename, "r");
    if (pf == NULL) {
        fprintf(stderr, "\nError in load_sorting_from_file: file \'%s\' not found\n", filename);
        exit(1);
    }
    //--- 
    for (int i = 0; i < D; ++i) {
        fscanf(pf, "%d ", &(idx_cluster_sort[i]));
    }
    fclose(pf);

    //print out
    fprintf(stderr, "\nLoading sorting from file \'%s\': ", filename);
    for (int i = 0; i < D; ++i) {
        fprintf(stderr, "%d ", idx_cluster_sort[i]);
    }
    fprintf(stderr, "\n");

    return 0;
}
//----------------------------------------------------------------------------------------------------------------------

int EMP::build_adjacence_matrix(parameters& param) {

    if (Np <= 0) {
        fprintf(stderr, "\nError in EMP::build_adjacence_matrix. Np=%d not valid.\n", Np);
        exit(1);
    }
    double **distance_matrix = new double*[N];
    for (int i = 0; i < N; i++) {
        distance_matrix[i] = new double[N];
        for (int j = 0; j < N; j++) {
            distance_matrix[i][j] = 0.0;
        }
    }
    double *copy_dist_column = new double[N];
    for (int j = 0; j < N; j++) {
        copy_dist_column[j] = 0.0;
    }
    //----

    for (int ii = 0; ii < N; ii++) {
        for (int jj = ii + 1; jj < N; jj++) {
            double r_tmp = 0.0;
            for (int pp = 0; pp < D; pp++) {
                r_tmp += fabs(data_matrix[ii][pp] - data_matrix[jj][pp]);
            }
            distance_matrix[ii][jj] = r_tmp;
            distance_matrix[jj][ii] = r_tmp;
        }
        distance_matrix[ii][ii] = 100000.0;
        for (int jj = 0; jj < N; jj++) {
            copy_dist_column[jj] = distance_matrix[ii][jj];
        }

        for (int pp = 0; pp < Np; pp++) {
            float min_tmp = 100000000.0;
            int i_min = 0;
            for (int jj = 0; jj < N; jj++) {
                if (copy_dist_column[jj] < min_tmp) {
                    min_tmp = copy_dist_column[jj];
                    i_min = jj;
                }
            }
            copy_dist_column[i_min] = 100000.0;
            adj_matrix[ii][i_min] = 1;
        }
    }

    for (int i = 0; i < N; i++)
        delete [] distance_matrix[i];
    delete[] distance_matrix;

    delete[] copy_dist_column;
    return 0;
}
/* *************************************************************************************************** */

int EMP::solve_convex(double& g, parameters& param) {

    reset_out_data();

    if (strcmp(param.clustering_type, "L1") == 0) {
        solve_L1(g);
    } 
    else if (strcmp(param.clustering_type, "EMDL1") == 0) {
        solve_EMDL1(g);
    } 
    else {
        fprintf(stderr, "\nError, clustering type \'%s\' not found\n", param.clustering_type);
        fprintf(stderr, "\nClustering type valid: L1, EMDL1. \n\n", param.clustering_type);
        exit(1);
    }
    fprintf(stderr, "Solve_convex done.\n");

    get_id_prototype();

    return 0;
}

/* *************************************************************************************************** */

int EMP::solve_L1(double& g) {

    double gamma_curr = g;
    // create linear programming problem
    int zsize = N*Np;
    int fsize = N*D;
    int psize = N*D;
    int n_var = psize + fsize + zsize;
    int n_cons = 2 * D * zsize + N + 2 * N*D;
    fprintf(stderr, "\nL1: Solving a problem with %d variables and %d constraints (zsize= %d)\n", n_var, n_cons, zsize);

    glp_prob *lp;
    int *ia = new int[n_cons * n_var + 1];
    int *ja = new int[n_cons * n_var + 1];
    double *ar = new double[n_cons * n_var + 1];
    
    lp = glp_create_prob();
    glp_set_prob_name(lp, "clustering L1");
    glp_set_obj_dir(lp, GLP_MIN);

    // specify equality constraints bounds
    double *da = new double[n_cons];
    for (int ii = 0; ii < n_cons; ii++) {
        da[ii] = 0.0;
    }
    for (int ii = 0; ii < N; ii++) {
        da[2 * zsize * D + ii] = 1.0;
    }

    int pa = 0;
    int qa = 0;
    for (int ii = 0; ii < N * D; ii++) {
        da[2 * zsize * D + N + ii] = -data_matrix[pa][qa];
        da[2 * zsize * D + N + N * D + ii] = data_matrix[pa][qa];
        //fprintf(stderr, "%d %d %f \n",pa, qa,data_matrix[pa][qa]);
        qa++;
        if (qa == D) {
            pa++;
            qa = 0;
        }
    }

    glp_add_rows(lp, n_cons);

    for (int ii = 1; ii <= n_cons; ii++) {
        glp_set_row_bnds(lp, ii, GLP_UP, 0.0, da[ii - 1]);
        //  fprintf(stderr, "%f \n", da[ii-1]);
    }
    for (int ii = 2 * zsize * D + 1; ii <= 2 * zsize * D + N; ii++) {
        glp_set_row_bnds(lp, ii, GLP_FX, da[ii - 1], 0.0);
        //fprintf(stderr, "%f \n", da[ii-1]);
    }
    delete [] da;

    // add lower bounds and objective functions
    double *d = new double[n_var];
    for (int ii = 0; ii < n_var; ii++) {
        d[ii] = 0.0;
    }
    for (int ii = 0; ii < fsize; ii++) {
        d[ii] = 1.0;
    }
    int ind = fsize + psize;
    for (int ii = 0; ii < zsize; ii++) {
        d[ind + ii] = gamma_curr;
    }
    glp_add_cols(lp, n_var);
    for (int ii = 1; ii <= n_var; ii++) {
        glp_set_col_bnds(lp, ii, GLP_LO, 0.0, 0.0);
        glp_set_obj_coef(lp, ii, d[ii - 1]);
        // fprintf(stderr, "%f \n", d[ii-1]);
    }
    delete [] d;

    // add other constraints
    pa = 1;
    qa = 1;
    int k = 1;
    int tt = 0;
    //first type - clustering constraints
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < N; jj++) {
            if (adj_matrix[ii][jj] > 0) {
                for (int kk = 0; kk < D; kk++) {
                    for (int pp = 0; pp < fsize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        k++;
                        qa++;
                    }
                    for (int pp = fsize; pp < fsize + psize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + ii * D + kk)) {
                            ar[k] = 1.0;
                        }
                        if (pp == (fsize + jj * D + kk)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    for (int pp = fsize + psize; pp < n_var; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + psize + tt)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    pa++;
                    qa = 1;
                }
                tt++;
            }
        }
    }
    tt = 0;
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < N; jj++) {
            if (adj_matrix[ii][jj] > 0) {
                for (int kk = 0; kk < D; kk++) {
                    for (int pp = 0; pp < fsize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        k++;
                        qa++;
                    }
                    for (int pp = fsize; pp < fsize + psize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + (ii) * D + kk)) {
                            ar[k] = -1.0;
                        }
                        if (pp == (fsize + jj * D + kk)) {
                            ar[k] = 1.0;
                        }
                        // fprintf(stderr, "%f ", ar[k]);
                        qa++;
                        k++;
                    }
                    for (int pp = fsize + psize; pp < n_var; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + psize + tt)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    pa++;
                    qa = 1;
                }
                tt++;
            }
        }
    }
    // multiple equality constraint for prototypes
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < n_var; jj++) {
            ia[k] = pa;
            ja[k] = qa;
            if ((jj >= fsize) && (jj < fsize + psize)) {
                if (floor((jj - fsize) / D) == ii) {
                    ar[k] = 1.0;

                } else {
                    ar[k] = 0.0;
                }
            } else {
                ar[k] = 0.0;
            }
            k++;
            qa++;
        }
        pa++;
        qa = 1;
    }
    tt = 0;
    //first type - l1 for histograms
    for (int ii = 1; ii <= N; ii++) {
        for (int kk = 0; kk < D; kk++) {
            for (int pp = 0; pp < fsize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                if (pp == ((ii - 1) * D + kk)) {
                    ar[k] = -1.0;
                }
                k++;
                qa++;
            }
            for (int pp = fsize; pp < psize + fsize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                if (pp == (fsize + (ii - 1) * D + kk)) {
                    ar[k] = -1.0;
                }
                qa++;
                k++;
            }
            for (int pp = fsize + psize; pp < n_var; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                qa++;
                k++;
            }
            pa++;
            qa = 1;
        }
        tt++;
    }
    tt = 0;
    //second type - l1 for histograms
    for (int ii = 1; ii <= N; ii++) {
        for (int kk = 0; kk < D; kk++) {
            for (int pp = 0; pp < fsize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                if (pp == ((ii - 1) * D + kk)) {
                    ar[k] = -1.0;
                }
                k++;
                qa++;
            }
            for (int pp = fsize; pp < psize + fsize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                if (pp == (fsize + (ii - 1) * D + kk)) {
                    ar[k] = 1.0;
                }
                qa++;
                k++;
            }
            for (int pp = fsize + psize; pp < n_var; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                qa++;
                k++;
            }
            pa++;
            qa = 1;
        }
        tt++;
    }
    glp_load_matrix(lp, n_cons*n_var, ia, ja, ar);
    delete[] ja;
    delete[] ia;
    delete[] ar;
    /* ************************************************************************************************************ */

    // SOLVE
    glp_simplex(lp, NULL);
    double z = glp_get_obj_val(lp);
    /* ************************************************************************************************************ */

    // load results into matrix
    int ic = 0;
    int iw = 0;
    for (int ii = fsize; ii < fsize + psize; ii++) {
        double x1 = glp_get_col_prim(lp, ii + 1);
        clustered_data_matrix[ic][iw] = x1;
        //fprintf(stderr, "%.2f ", x1);
        ++iw;
        if (((ii + 1) % D) == 0.0) {
            ++ic;
            iw = 0;
            // fprintf(stderr, "\n[%d]", ic);
        }
    }
    glp_delete_prob(lp);
}

/* *************************************************************************************************** */
int EMP::solve_EMDL1(double& g) {

    double gamma_curr = g;

    // create linear programming problem
    int zsize = N*Np;
    int n_var = 2 * N * (D - 1) + N * D + zsize;
    int n_cons = 2 * D * zsize + N + N*D;
    int fsize = 2 * N * (D - 1);
    int psize = N*D;
    fprintf(stderr, "EMD_L1: Solving a problem with %d variables and %d constraints (zsize= %d)\n", n_var, n_cons, zsize);

    glp_prob *lp;
    int *ia = new int[n_cons * n_var + 1];
    int *ja = new int[n_cons * n_var + 1];
    double *ar = new double[n_cons * n_var + 1];
  
    lp = glp_create_prob();
    glp_set_prob_name(lp, "clustering EMD");
    glp_set_obj_dir(lp, GLP_MIN);

    // specify equality constraints bounds
    double *da = new double[n_cons];
    for (int ii = 0; ii < n_cons; ii++) {
        da[ii] = 0.0;
    }
    for (int ii = 0; ii < N; ii++) {
        da[2 * zsize * D + ii] = 1.0;
    }
    int pa = 0;
    int qa = 0;
    for (int ii = 0; ii < N * D; ii++) {
        da[2 * zsize * D + N + ii] = data_matrix[pa][qa];
        qa++;
        if (qa == D) {
            pa++;
            qa = 0;
        }
    }
    glp_add_rows(lp, n_cons);

    for (int ii = 1; ii <= n_cons; ii++) {
        glp_set_row_bnds(lp, ii, GLP_UP, 0.0, da[ii - 1]);
    }
    for (int ii = 2 * zsize * D + 1; ii <= n_cons; ii++) {
        glp_set_row_bnds(lp, ii, GLP_FX, da[ii - 1], 0.0);
    }
    delete [] da;

    // add lower bounds and objective functions
    double *d = new double[n_var];
    for (int ii = 0; ii < n_var; ii++) {
        d[ii] = 0.0;
    }
    int k = 0;
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < 2 * (D - 1); jj++) {
            d[k] = 1.0;
            k++;
        }
    }
    int ind = fsize + psize;
    for (int ii = 0; ii < zsize; ii++) {
        d[ind + ii] = gamma;
    }
    glp_add_cols(lp, n_var);
    for (int ii = 1; ii <= n_var; ii++) {
        glp_set_col_bnds(lp, ii, GLP_LO, 0.0, 0.0);
        glp_set_obj_coef(lp, ii, d[ii - 1]);
        // printf("%f \n", d[ii-1]);
    }
    delete[] d;

    // add other constraints
    pa = 1;
    qa = 1;
    k = 1;
    int tt = 0;
    //first type - clustering constraints
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < N; jj++) {
            if (adj_matrix[ii][jj] > 0) {
                for (int kk = 0; kk < D; kk++) {
                    for (int pp = 0; pp < fsize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        k++;
                        qa++;
                    }
                    for (int pp = fsize; pp < fsize + psize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + ii * D + kk)) {
                            ar[k] = 1.0;
                        }
                        if (pp == (fsize + jj * D + kk)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    for (int pp = fsize + psize; pp < n_var; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + psize + tt)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    pa++;
                    qa = 1;
                }
                tt++;
            }
        }
    }
    tt = 0;
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < N; jj++) {
            if (adj_matrix[ii][jj] > 0) {
                for (int kk = 0; kk < D; kk++) {
                    for (int pp = 0; pp < fsize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        k++;
                        qa++;
                    }
                    for (int pp = fsize; pp < fsize + psize; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + (ii) * D + kk)) {
                            ar[k] = -1.0;
                        }
                        if (pp == (fsize + jj * D + kk)) {
                            ar[k] = 1.0;
                        }
                        qa++;
                        k++;
                    }
                    for (int pp = fsize + psize; pp < n_var; pp++) {
                        ia[k] = pa;
                        ja[k] = qa;
                        ar[k] = 0.0;
                        if (pp == (fsize + psize + tt)) {
                            ar[k] = -1.0;
                        }
                        qa++;
                        k++;
                    }
                    pa++;
                    qa = 1;
                }
                tt++;
            }
        }
    }
    // multiple equality constraint for prototypes
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < n_var; jj++) {
            ia[k] = pa;
            ja[k] = qa;
            if ((jj >= fsize) && (jj < fsize + psize)) {
                if (floor((jj - fsize) / D) == ii) {
                    ar[k] = 1.0;
                } else {
                    ar[k] = 0.0;
                }
            } else {
                ar[k] = 0.0;
            }
            k++;
            qa++;
        }
        pa++;
        qa = 1;
    }
    // inequality constraint tipo a
    // inequality constraint tipo b
    for (int ii = 0; ii < N; ii++) {
        for (int jj = 0; jj < D; jj++) {
            for (int pp = 0; pp < fsize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                int ttt = floor(pp / (2 * (D - 1)));
                if (ii == ttt) {
                    int rem = (pp % (2 * (D - 1)));
                    if ((jj == 0) && (rem == 0))
                        ar[k] = 1.0;
                    if ((jj == 0) && (rem == (D - 1)))
                        ar[k] = -1.0;
                    if ((jj == D - 1) && (rem == (D - 2)))
                        ar[k] = -1.0;
                    if ((jj == D - 1) && (rem == (2 * (D - 1) - 1)))
                        ar[k] = 1.0;
                    if ((jj > 0) && (jj < D - 1)) {
                        if (rem < D - 1) {
                            if (rem == jj - 1)
                                ar[k] = -1.0;
                            if ((rem) == jj)
                                ar[k] = 1.0;
                        } else {
                            if ((rem - D + 1) == (jj - 1))
                                ar[k] = 1.0;
                            if ((rem - D + 1) == jj)
                                ar[k] = -1.0;
                        }
                    }
                }
                k++;
                qa++;
            }
            for (int pp = fsize; pp < fsize + psize; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                if (floor((pp - fsize)) == (ii * D + jj)) {
                    ar[k] = 1.0;
                }
                qa++;
                k++;
            }
            for (int pp = fsize + psize; pp < n_var; pp++) {
                ia[k] = pa;
                ja[k] = qa;
                ar[k] = 0.0;
                qa++;
                k++;
            }
            pa++;
            qa = 1;
        }
    }
    glp_load_matrix(lp, n_cons*n_var, ia, ja, ar);
    delete[] ja;
    delete[] ia;
    delete[] ar;

    /* ************************************************************************************************************ */
    // SOLVE
    glp_simplex(lp, NULL);
    double z = glp_get_obj_val(lp);
    /* ************************************************************************************************************ */

    int nn = 0;
    int dd = 0;
    for (int ii = fsize; ii < fsize + psize; ii++) {
        double x1 = glp_get_col_prim(lp, ii + 1);
        clustered_data_matrix[nn][dd] = x1;
        // fprintf(stderr, "%.2lf ", x1);
        ++dd;
        if (((ii - fsize + 1) % D) == 0) {
            ++nn;
            dd = 0;
            // fprintf(stderr, "\n[%d]", nn);
        }
    }
    glp_delete_prob(lp);
}
//--------------------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------------------

int EMP::get_id_prototype() {

    n_prot = 0;
    //
    double** unique_prototypes = new double*[N];
    for (int i = 0; i < N; ++i) {
        unique_prototypes[i] = new double[D];
        memset(unique_prototypes[i], 0, D * sizeof (double));
    }
    //---
    double eps = 0.000001;
    //---
    for (int i = 0; i < N; ++i) {

        int cluster_id = -1;
        //
        for (int ip = 0; ip < n_prot; ++ip) {
            double dist2prototype = dist_L1(clustered_data_matrix[i], unique_prototypes[ip], D);
            //fprintf(stderr, "\ndist ( cl[%d], p[%d] ) -> %f", i, ip, dist2prototype);
            if (dist2prototype < eps) {
                cluster_id = ip;
                break;
            }
        }
        if (cluster_id == -1) {
            //add found new prototype to list
            cluster_id = n_prot;
            memcpy(unique_prototypes[cluster_id], clustered_data_matrix[i], D * sizeof (double));
            ++n_prot;
        }
        prototype_idx[i] = cluster_id;
    }
    //save_results_p(clustered_data_matrix, prototypes, n_prot, gamma, param);

    for (int i = 0; i < N; ++i) {
        delete[] unique_prototypes[i];
    }
    delete unique_prototypes;

}
//--------------------------------------------------------------------------------------------------------------

void EMP::print_adj_matrix() {

    fprintf(stderr, "\n");
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            fprintf(stderr, " %d", adj_matrix[i][j]);
        }
        fprintf(stderr, "\n");
    }
}
//--------------------------------------------------------------------------------------------------------------

void EMP::save_results(parameters& param) {

    // print to screen
    fprintf(stderr, "\nprototypes found = %d \n\n", n_prot);
    fprintf(stderr, "final clustering results:\n");
    for (int i = 0; i < N; ++i) {
        fprintf(stderr, "%d ", prototype_idx[i]);
    }
    fprintf(stderr, "\n\n");
    //---

    // save to file
    char filename[300];
    sprintf(filename, "%s/%s", param.output_dir, param.file_res_clustering);
    FILE* pf = fopen(filename, "w+");
    if (pf == NULL) {
        fprintf(stderr, "\nError in save_results. impossible top open file \'%s\' for writing. \n", filename);
        exit(1);
    }
    for (int i = 0; i < N; ++i) {
        fprintf(pf, "%d ", prototype_idx[i]);
    }
    fprintf(pf, "\n");
    fclose(pf);
}
//--------------------------------------------------------------------------------------------------------------

void EMP::delete_data() {

    delete[] idx_cluster_sort;
    //---
    for (int i = 0; i < N; i++)
        delete[] data_matrix[i];
    delete[] data_matrix;
    //---
    for (int i = 0; i < N; i++)
        delete[] adj_matrix[i];
    delete[] adj_matrix;
    //---    
    if (clustered_data_matrix != NULL) {
        for (int i = 0; i < N; i++)
            delete[] clustered_data_matrix[i];
        delete[] clustered_data_matrix;
    }
    //---
    delete[] prototype_idx;
}
//----------------------------------------------------------------------------------------------------------------------
/* *************************************************************************************************** */

double dist_L1(double* vec1, double* vec2, int& D) {
    double dist = 0.0;
    for (int i = 0; i < D; ++i) {
        dist += fabs(vec1[i] - vec2[i]);
    }
    return dist;
}
/* *************************************************************************************************** */


